import re
import subprocess

# for macos
class SwapUsageReader:
    def __init__(self):
        pass

    def parse_memstr(self, size_str):
        match = re.match(r"(\d+(\.\d+)?)([KMG]?)", size_str)
        if not match:
            raise ValueError("Invalid memory size format")
        number, _, unit = match.groups()
        number = float(number)

        if unit == "G":
            return number * 1024 * 1024 * 1024
        elif unit == "M":
            return number * 1024 * 1024
        elif unit == "K":
            return number * 1024
        else:
            return number

    def read(self):
        res = {}
        try:
            swap_stats = subprocess.run(["sysctl", "vm.swapusage"], capture_output=True, text=True)
            tokens = swap_stats.stdout.strip().split(' ')
            res['swap used'] = self.parse_memstr(tokens[7])
        except:
            # log something
            pass
        return res


import psutil
import time

from cubestat.readers.mem_reader import MemReader
from cubestat.readers.nv_reader import NVReader
from cubestat.readers.free_swap_reader import FreeSwapReader

class LinuxReader:
    def __init__(self, interval_ms):
        self.first = True
        self.interval_ms = interval_ms
        self.mem_reader = MemReader(interval_ms)
        self.nv = NVReader()
        self.swap_reader = FreeSwapReader()

    def read(self):
        res = self.mem_reader.read()
        res['swap'] = self.swap_reader.read()

        disk_load = psutil.disk_io_counters()
        nw_load = psutil.net_io_counters()
        d = self.interval_ms / 1000.0

        # TODO: numa nodes here?
        cpu_clusters = []
        cpu_load = psutil.cpu_percent(percpu=True)

        cluster_title = f'[{len(cpu_load)}] Total CPU Util, %'
        cpu_clusters.append(cluster_title)
        total_load = 0.0
        res['cpu'][cluster_title] = 0.0

        for i, v in enumerate(cpu_load):
            title = f'CPU {i} util %'
            res['cpu'][title] = v
            total_load += v
        res['cpu'][cluster_title] = total_load / len(cpu_load)

        res['gpu'] = self.nv.read()

        if self.first:
            self.disk_read_last = disk_load.read_bytes
            self.disk_written_last = disk_load.write_bytes
            self.network_read_last = nw_load.bytes_recv
            self.network_written_last = nw_load.bytes_sent
            self.first = False

        res['disk']['disk read'] = ((disk_load.read_bytes - self.disk_read_last) / d)
        res['disk']['disk write'] = ((disk_load.write_bytes - self.disk_written_last) / d)
        self.disk_read_last = disk_load.read_bytes
        self.disk_written_last = disk_load.write_bytes

        res['network']['network rx'] = ((nw_load.bytes_recv - self.network_read_last) / d)
        res['network']['network tx'] = ((nw_load.bytes_sent - self.network_written_last) / d)
        self.network_read_last = nw_load.bytes_recv
        self.network_written_last = nw_load.bytes_sent

        return res.items(), cpu_clusters
    
    def loop(self, on_snapshot_cb):
        # TODO: should this be monotonic?
        begin_ts = time.time()
        n = 0
        d = self.interval_ms / 1000.0
        while True:
            snapshot, cpu_clusters = self.read()
            on_snapshot_cb(snapshot, cpu_clusters)
            n += 1
            expected_time = begin_ts + n * d
            current_time = time.time()
            if expected_time > current_time:
                time.sleep(expected_time - current_time)



import psutil

# based on psutil
class MemReader:
    def __init__(self, interval_ms):
        self.first = True
        self.interval_ms = interval_ms

    # TODO: make this return only memory
    def read(self):
        d = self.interval_ms / 1000.0
        res = {
            'cpu': {},
            'ram': {'RAM used %': psutil.virtual_memory().percent},
            'swap': {},
            'gpu': {},
            'ane': {},
            'disk': {},
            'network': {},
            'power': {},
        }
        return res


import subprocess
import plistlib

from cubestat.readers.mem_reader import MemReader
from cubestat.readers.swapusage_reader import SwapUsageReader

# reading from powermetrics
class AppleReader:
    # This is pretty much a guess based on tests on a few models I had available.
    # Need anything M3 + Ultra models to test.
    # Based on TOPS numbers Apple published, all models seem to have some ANE 
    # except Ultra having 2x.
    ane_power_scalers = {
        "M1": 13000.0,
        "M2": 15500.0,
        "M3": 15500.0,
    }

    def __init__(self, interval_ms) -> None:
        # identity the model to get ANE scaler
        brand_str = subprocess.check_output(['sysctl', '-n', 'machdep.cpu.brand_string'], text=True)
        self.ane_scaler = 15500 # default to M2
        for k, v in self.ane_power_scalers.items():
            if k in brand_str:
                self.ane_scaler = v
                if 'ultra' in brand_str.lower():
                    self.ane_scaler *= 2
                break
        
        cmd = ['sudo', 'powermetrics', '-f', 'plist', '-i', str(interval_ms), '-s', 'cpu_power,gpu_power,ane_power,network,disk']
        self.powermetrics = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # we are getting first line here to allow user to enter sudo credentials before 
        # curses initialization.
        self.firstline = self.powermetrics.stdout.readline()

        self.mem_reader = MemReader(interval_ms)
        self.swap_reader = SwapUsageReader()

    def read(self, snapshot):
        res = self.mem_reader.read()
        res['swap'] = self.swap_reader.read()

        cpu_clusters = []
        for cluster in snapshot['processor']['clusters']:
            idle_cluster, total_cluster = 0.0, 0.0
            n_cpus = len(cluster['cpus'])
            cluster_title = f'[{n_cpus}] {cluster["name"]} total CPU util %'
            cpu_clusters.append(cluster_title)
            res['cpu'][cluster_title] = 0.0
            for cpu in cluster['cpus']:
                title = f'{cluster["name"]} CPU {cpu["cpu"]} util %'
                res['cpu'][title] = 100.0 - 100.0 * cpu['idle_ratio']
                idle_cluster += cpu['idle_ratio']
                total_cluster += 1.0
            res['cpu'][cluster_title] = 100.0 - 100.0 * idle_cluster / total_cluster

        res['gpu']['GPU util %'] = 100.0 - 100.0 * snapshot['gpu']['idle_ratio']
        
        res['ane']['ANE util %'] = 100.0 * snapshot['processor']['ane_power'] / self.ane_scaler

        res['power']['total power'] = snapshot['processor']['combined_power']
        res['power']['ANE power'] = snapshot['processor']['ane_power']
        res['power']['CPU power'] = snapshot['processor']['cpu_power']
        res['power']['GPU power'] = snapshot['processor']['gpu_power']

        res['disk']['disk read'] = snapshot['disk']['rbytes_per_s']
        res['disk']['disk write'] = snapshot['disk']['wbytes_per_s']
        res['network']['network rx'] = snapshot['network']['ibyte_rate']
        res['network']['network tx'] = snapshot['network']['obyte_rate']
        return res.items(), cpu_clusters
    
    def loop(self, on_snapshot_cb):
        buf = bytearray()
        buf.extend(self.firstline)

        while True:
            line = self.powermetrics.stdout.readline()
            buf.extend(line)
            # we check for </plist> rather than '0x00' because powermetrics injects 0x00 
            # right before the measurement event, not right after. So, if we were to wait 
            # for 0x00 we'll be delaying next sample by sampling period. 
            if b'</plist>\n' == line:
                snapshot, cpu_clusters = self.read(plistlib.loads(bytes(buf).strip(b'\x00')))
                on_snapshot_cb(snapshot, cpu_clusters)
                buf.clear()


import subprocess
from importlib.util import find_spec

# reading compute/memory utilization for nvidia cards. 
class NVReader:
    def __init__(self):
        self.has_nvidia = False
        try:
            subprocess.check_output('nvidia-smi')
            nvspec = find_spec('pynvml')
            if nvspec is not None:
                from pynvml.smi import nvidia_smi
                self.nvsmi = nvidia_smi.getInstance()
                self.has_nvidia = True
        except Exception:
            # TODO: add logging here
            pass

    def read(self):
        res = {}
        total = 0
        if self.has_nvidia:
            n_gpus = 0
            for i, v in enumerate(self.nvsmi.DeviceQuery('utilization.gpu,memory.total,memory.used')['gpu']):
                res[f'GPU {i} util %'] = v['utilization']['gpu_util']
                total += v['utilization']['gpu_util']
                res[f'GPU {i} vram used %'] = 100.0 * v['fb_memory_usage']['used'] / v['fb_memory_usage']['total']
                n_gpus += 1
            if n_gpus > 1:
                combined = {}
                combined[f'[{n_gpus}] Total GPU util %'] = total / n_gpus
                for k, v in res.items():
                    combined[k] = v
                return combined
        return res


#!/usr/bin/env python3

import argparse
import collections
import curses
import itertools
import os
import sys
import math

from enum import Enum
from math import floor
from threading import Thread, Lock

from cubestat.readers.linux_reader import LinuxReader
from cubestat.readers.macos_reader import AppleReader

class EnumLoop(Enum):
    def next(self):
        values = list(self.__class__)
        return values[(values.index(self) + 1) % len(values)]
    
class EnumStr(Enum):
    def __str__(self):
        return self.value

class Legend(EnumLoop, EnumStr):
    hidden = 'off'
    last = 'last'
    
class CPUMode(EnumLoop, EnumStr):
    all = 'all'
    by_cluster = 'by_cluster'
    by_core = 'by_core'

class PowerMode(EnumLoop, EnumStr):
    combined = 'combined'
    all = 'all'
    off = 'off'

class GPUMode(EnumLoop, EnumStr):
    collapsed = 'collapsed'
    load_only = 'load_only'
    load_and_vram = 'load_and_vram'
    
class Color(EnumStr):
    red = 'red'
    green = 'green'
    blue = 'blue'
    pink = 'pink'
    olive = 'olive'
    navy = 'navy'
    blue_dark = 'blue_dark'
    purple = 'purple'
    mixed = 'mixed'
    dark = 'dark'

def auto_cpu_mode():
     return CPUMode.all if os.cpu_count() < 40 else CPUMode.by_cluster

parser = argparse.ArgumentParser("cubestat")
parser.add_argument('--refresh_ms', '-i', type=int, default=1000, help='Update frequency, milliseconds')
parser.add_argument('--buffer_size', type=int, default=500, help='How many datapoints to store. Having it larger than screen width is a good idea as terminal window can be resized')
parser.add_argument('--cpu', type=CPUMode, default=auto_cpu_mode(), choices=list(CPUMode), help='CPU mode - showing all cores, only cumulative by cluster or both. Can be toggled by pressing c.')
parser.add_argument('--gpu', type=GPUMode, default=GPUMode.load_only, choices=list(GPUMode), help='GPU mode - hidden, showing all GPUs load, or showing load and vram usage. Can be toggled by pressing g.')
parser.add_argument('--power', type=PowerMode, default=PowerMode.combined, choices=list(PowerMode), help='Power mode - off, showing breakdown CPU/GPU/ANE load, or showing combined usage. Can be toggled by pressing p.')
parser.add_argument('--color', type=Color, default=Color.mixed, choices=list(Color))
parser.add_argument('--legend', type=Legend, default=Legend.last, choices=list(Legend), help='Show/hide numeric utilization percentage. Can be toggled by pressing l.')
parser.add_argument('--disk', action="store_true", default=True, help="Show disk read/write. Can be toggled by pressing d.")
parser.add_argument('--swap', action="store_true", default=True, help="Show swap . Can be toggled by pressing s.")
parser.add_argument('--network', action="store_true", default=True, help="Show network io. Can be toggled by pressing n.")
parser.add_argument('--no-disk', action="store_false", dest="disk", help="Hide disk read/write. Can be toggled by pressing d.")
parser.add_argument('--no-swap', action="store_false", default=True, help="Hide swap. Can be toggled by pressing s.")
parser.add_argument('--no-network', action="store_false", dest="network", help="Hide network io. Can be toggled by pressing n.")

args = parser.parse_args()

class Horizon:
    def __init__(self, stdscr, reader):
        stdscr.nodelay(False)
        stdscr.timeout(50)
        curses.curs_set(0)
        curses.start_color()
        curses.use_default_colors()

        self.spacing_width = 1
        self.filling = '.'

        self.cells = self.prepare_cells()
        self.stdscr = stdscr

        # all of the fields below are mutable and can be accessed from 2 threads
        self.lock = Lock()
        self.data = {k: collections.defaultdict(lambda: collections.deque(maxlen=args.buffer_size)) for k in ['cpu', 'ram', 'swap', 'gpu', 'ane', 'disk', 'network', 'power']}
        dark_colormap = {
            'cpu': Color.purple,
            'ram': Color.navy,
            'gpu': Color.blue_dark,
            'ane': Color.blue_dark,
            'disk': Color.olive,
            'network': Color.olive,
            'swap': Color.navy,
            'power': Color.blue_dark,
        }
        light_colormap = {
            'cpu': Color.green,
            'ram': Color.pink,
            'gpu': Color.red,
            'ane': Color.red,
            'disk': Color.blue,
            'network': Color.blue,
            'swap': Color.pink,
            'power': Color.red,
        }

        if args.color == Color.mixed:
            self.colormap = light_colormap
        elif args.color == Color.dark:
            self.colormap = dark_colormap
        else:
            self.colormap = {k: args.color for k, _ in light_colormap.items()}
        self.snapshots_observed = 0
        self.snapshots_rendered = 0
        self.legend_mode = args.legend
        self.cpumode = args.cpu
        self.show_disk = args.disk
        self.show_swap = args.swap
        self.show_network = args.network
        self.power_mode = args.power
        self.gpumode = args.gpu
        self.settings_changed = False
        self.reader = reader
        self.vertical_shift = 0
        self.horizontal_shift = 0

    def prepare_cells(self):
        chrs = [' ', '▁', '▂', '▃', '▄', '▅', '▆', '▇', '█']
        colorschemes = {
            Color.green: [-1, 150, 107, 22],
            Color.red: [-1, 224, 181, 138],
            Color.blue: [-1, 189, 146, 103],
            Color.pink: [-1, 223, 180, 137],
            Color.olive: [-1, 58, 101, 144],
            Color.navy: [-1, 18, 61, 105],
            Color.blue_dark: [-1, 23, 66, 109],
            Color.purple: [-1, 53, 96, 138]
        }
        cells = {}
        colorpair = 1
        for name, colors in colorschemes.items():
            cells[name] = []
            for fg, bg in zip(colors[1:], colors[:-1]):
                curses.init_pair(colorpair, fg, bg)
                cells[name].extend((chr, colorpair) for chr in chrs)
                colorpair += 1
        return cells

    def process_snapshot(self, data, cpu_clusters):
        for group, vals in data:
            for title, value in vals.items():
                with self.lock:
                    self.data[group][title].append(value)

        with self.lock:
            self.cpu_clusters = cpu_clusters
            self.snapshots_observed += 1
            if self.horizontal_shift > 0:
                self.horizontal_shift += 1

    def write_string(self, row, col, s, color=0):
        if row < 0 or row >= self.rows or col < 0:
            return
        if col + len(s) > self.cols:
            s = s[:self.cols - col]
        try:
            self.stdscr.addstr(row, col, s, color)
        except:
            # TODO: log something
            pass

    def write_char(self, row, col, chr, color=0):
        if row < 0 or row >= self.rows or col < 0 or col >= self.cols:
            return
        try:
            self.stdscr.addch(row, col, chr, color)
        except:
            # TODO: log something
            pass

    # buckets is a list of factor/label, e.g. [(1024*1024, 'Mb'), (1024, 'Kb'), (1, 'b')]
    def format_measurement(self, spacing, curr, mx, buckets):
        if self.legend_mode != Legend.last:
            return f'{spacing}╗'
        for lim, unit in buckets[:-1]:
            if mx > lim:
                return f'last:{curr / lim :3.0f}|{int(mx / lim)}{unit}{spacing}╗'
        return f'last:{curr :3.0f}|{int(mx)}{buckets[-1][1]}{spacing}╗'

    def render(self):
        with self.lock:
            if self.snapshots_observed == self.snapshots_rendered and not self.settings_changed:
                return
        self.stdscr.erase()
        self.rows, self.cols = self.stdscr.getmaxyx()
        spacing = ' ' * self.spacing_width

        # Each chart takes two lines, with format roughly
        # ╔ GPU util %........................................................................last:  4% ╗
        # ╚ ▁▁▁  ▁    ▁▆▅▄ ▁▁▁      ▂ ▇▃▃▂█▃▇▁▃▂▁▁▂▁▁▃▃▂▁▂▄▄▁▂▆▁▃▁▂▃▁▁▁▂▂▂▂▂▂▁▁▃▂▂▁▂▁▃▄▃ ▁▁▃▁▄▂▃▂▂▂▃▃▅▅ ╝

        with self.lock:
            i = 0
            skip = self.vertical_shift
            is_multigpu = len(self.data['gpu']) > 1
            for group_name, group in self.data.items():
                if group_name == 'disk' and not self.show_disk:
                    continue
                if group_name == 'network' and not self.show_network:
                    continue
                if group_name == 'swap' and not self.show_swap:
                    continue

                cells = self.cells[self.colormap[group_name]]
                range = len(cells)
                for title, series in group.items():
                    indent = ''

                    if group_name == 'cpu':
                        if self.cpumode == CPUMode.by_cluster and title not in self.cpu_clusters:
                            continue
                        if self.cpumode == CPUMode.by_core and title in self.cpu_clusters:
                            continue
                        if self.cpumode == CPUMode.all and title not in self.cpu_clusters:
                            indent = '  '

                    if group_name == 'gpu':
                        if is_multigpu and self.gpumode == GPUMode.collapsed and "Total GPU" not in title:
                            continue
                        if self.gpumode == GPUMode.load_only and "vram" in title:
                            continue
                        if is_multigpu and "Total GPU" not in title:
                            indent = '  '

                    if group_name == 'power':
                        if self.power_mode == PowerMode.off:
                            continue
                        if self.power_mode == PowerMode.combined and 'total' not in title:
                            continue
                        if 'total' not in title:
                            indent = '  '
                    
                    if skip > 0:
                        skip -= 1
                        continue

                    # render title and left border, for example
                    #
                    # ╔ GPU util %
                    # ╚
                    titlestr = f'{indent}╔{spacing}{title}'
                    self.write_string(i * 2, 0, titlestr)
                    self.write_string(i * 2 + 1, 0, f'{indent}╚')

                    # data slice size
                    length = len(series) - self.horizontal_shift if self.horizontal_shift > 0 else len(series)
                    
                    # chart area width
                    width = self.cols - 2 * self.spacing_width - 2 - len(indent)
                    index = max(0, length - width)
                    data_slice = list(itertools.islice(series, index, min(index + width, length)))

                    # for percentage-like measurements
                    B = 100.0
                    strvalue = f'last:{data_slice[-1]:3.0f}%{spacing}╗' if self.legend_mode == Legend.last else f'{spacing}╗'
                    
                    if group_name == 'disk' or group_name == 'network':
                        B = max(data_slice)
                        B = float(1 if B == 0 else 2 ** (int((B - 1)).bit_length()))
                        strvalue = self.format_measurement(spacing, data_slice[-1], B, [(1024 * 1024, 'MB/s'), (1024, 'KB/s'), (1, 'Bytes/s')])

                    if group_name == 'swap':
                        B = max(data_slice)
                        B = float(1 if B == 0 else 2 ** (int((B - 1)).bit_length()))
                        strvalue = self.format_measurement(spacing, data_slice[-1], B, [(1024 * 1024, 'MB'), (1024, 'KB'), (1, 'Bytes')])

                    if group_name == 'power':
                        B = max(data_slice)
                        B = float(1 if B <= 0 else 10 ** math.ceil(math.log10(B)))
                        strvalue = self.format_measurement(spacing, data_slice[-1], B, [(1000 * 1000, 'kW'), (1000, 'W'), (1, 'mW')])

                    # render the rest of title row
                    #
                    # ╔ GPU util %........................................................................last:  4% ╗
                    # ╚
                    title_filling = self.filling * (self.cols - len(strvalue) - len(titlestr))
                    self.write_string(i * 2, len(titlestr), title_filling)
                    self.write_string(i * 2, self.cols - len(strvalue), strvalue)

                    # render the right border
                    #
                    # ╔ GPU util %........................................................................last:  4% ╗
                    # ╚                                                                                             ╝
                    border = f'{spacing}╝'
                    self.write_string(i * 2 + 1, self.cols - len(border), border)

                    # Render the chart itself
                    #
                    # ╔ GPU util %........................................................................last:  4% ╗
                    # ╚ ▁▁▁  ▁    ▁▆▅▄ ▁▁▁      ▂ ▇▃▃▂█▃▇▁▃▂▁▁▂▁▁▃▃▂▁▂▄▄▁▂▆▁▃▁▂▃▁▁▁▂▂▂▂▂▂▁▁▃▂▂▁▂▁▃▄▃ ▁▁▃▁▄▂▃▂▂▂▃▃▅▅ ╝
                    scaler = range / B
                    col = self.cols - (len(data_slice) + self.spacing_width) - 2
                    for v in data_slice:
                        col += 1
                        cell_index = floor(v * scaler)
                        if cell_index <= 0:
                            continue
                        if cell_index >= range:
                            cell_index = range - 1
                        chr, color_pair = cells[cell_index]
                        self.write_char(i * 2 + 1, col, chr, curses.color_pair(color_pair))

                    i += 1

                self.snapshots_rendered += 1
                
        self.stdscr.refresh()

    def render_loop(self):
        self.stdscr.keypad(True)
        while True:
            self.render()
            key = self.stdscr.getch()
            if key == ord('q') or key == ord('Q'):
                exit(0)
            if key == ord('l'):
                with self.lock:
                    self.legend_mode = self.legend_mode.next()
                    self.settings_changed = True
            if key == ord('c'):
                with self.lock:
                    self.cpumode = self.cpumode.next()
                    self.settings_changed = True
            if key == ord('p'):
                with self.lock:
                    self.power_mode = self.power_mode.next()
                    self.settings_changed = True
            if key == ord('g'):
                with self.lock:
                    self.gpumode = self.gpumode.next()
                    self.settings_changed = True
            if key == ord('s'):
                with self.lock:
                    self.show_swap = not self.show_swap
                    self.settings_changed = True
            if key == ord('d'):
                with self.lock:
                    self.show_disk = not self.show_disk
                    self.settings_changed = True
            if key == ord('n'):
                with self.lock:
                    self.show_network = not self.show_network
                    self.settings_changed = True
            if key == curses.KEY_UP:
                with self.lock:
                    if self.vertical_shift > 0:
                        self.vertical_shift -= 1
                        self.settings_changed = True
            if key == curses.KEY_DOWN:
                with self.lock:
                    self.vertical_shift += 1
                    self.settings_changed = True
            if key == curses.KEY_LEFT:
                with self.lock:
                    if self.horizontal_shift + 1 < self.snapshots_observed:
                        self.horizontal_shift += 1
                        self.settings_changed = True
            if key == curses.KEY_RIGHT:
                with self.lock:
                    if self.horizontal_shift > 0:
                        self.horizontal_shift -= 1
                        self.settings_changed = True
            if key == ord('0'):
                with self.lock:
                    if self.horizontal_shift > 0:
                        self.horizontal_shift = 0
                        self.settings_changed = True

    def loop(self):
        reader_thread = Thread(target=self.reader.loop, daemon=True, args=[self.process_snapshot])
        reader_thread.start()
        self.render_loop()

def start(stdscr, reader):
    h = Horizon(stdscr, reader)
    h.loop()

def main():
    if sys.platform == "darwin":
        curses.wrapper(start, AppleReader(args.refresh_ms))
    if sys.platform == "linux" or sys.platform == "linux2":
        curses.wrapper(start, LinuxReader(args.refresh_ms))
    # TODO: write something about platform not supported

if __name__ == '__main__':
    main()



import subprocess

class FreeSwapReader:
    def __init__(self):
        pass

    def read(self):
        res = {}
        try:
            swap_stats = subprocess.run(["free", "-b"], capture_output=True, text=True)
            lines = swap_stats.stdout.splitlines()
            for l in lines:
                if l.startswith("Swap:"):
                    parts = l.split()
                    res['swap used'] = float(parts[2])
        except:
            # TODO: log something
            pass

        return res
        